"use strict";

Object.defineProperty(exports, "__esModule", {
  value: true
});

var _types = require("../types");

var _types2 = _interopRequireDefault(_types);

var _astLocUtils = require("ast-loc-utils");

var _functions = require("../helpers/functions");

var _blocks = require("../helpers/blocks");

var _variables = require("../helpers/variables");

var _classes = require("../transforms/classes");

var _match = require("../transforms/match");

var matching = _interopRequireWildcard(_match);

var _comprehension = require("../transforms/comprehension");

var _safe = require("../transforms/safe");

var _inlinedOperators = require("../transforms/inlinedOperators");

var _for = require("../transforms/for");

var _catchExpression = require("../transforms/catchExpression");

var _stdlib = require("../state/stdlib");

function _interopRequireWildcard(obj) { if (obj && obj.__esModule) { return obj; } else { var newObj = {}; if (obj != null) { for (var key in obj) { if (Object.prototype.hasOwnProperty.call(obj, key)) newObj[key] = obj[key]; } } newObj.default = obj; return newObj; } }

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

// Main AST visitor
exports.default = function mainPass(compilerState, programPath) {
  var opts = compilerState.opts;
  // Main LSC transforms

  programPath.traverse({
    ForInArrayStatement(path) {
      // TODO: push the linter conditional down into the transform
      // and get linter status from compilerState
      if (opts.__linter) {
        (0, _for.lintForInArrayStatement)(path);
      } else {
        (0, _for.transformForInArrayStatement)(path);
      }
    }, ForInObjectStatement(path) {
      // TODO: push the linter conditional down into the transform
      // and get linter status from compilerState
      if (opts.__linter) {
        (0, _for.lintForInObjectStatement)(path);
      } else {
        (0, _for.transformForInObjectStatement)(path);
      }
    }, CallExpression(path) {
      var node = path.node;

      (0, _inlinedOperators.maybeReplaceWithInlinedOperator)(path, node.callee, node.arguments);
    }, NamedArrowFunction(path) {
      if (path.node.skinny) {
        (0, _functions.replaceWithPlainFunction)(path);
      } else if (path.node.generator) {
        // there are no arrow-generators in ES6, so can't compile to arrow
        (0, _functions.replaceWithBoundFunction)(path);
      } else {
        (0, _functions.replaceWithArrowFunction)(path);
      }
    }, NamedArrowMemberExpression(path) {
      var object = path.node.object;
      var node = path.node;
      delete node.object;
      node.type = "NamedArrowExpression";

      if (!node.skinny) {
        node.skinny = true; // binding here, don't turn into arrow
        node = (0, _astLocUtils.placeTreeAtLocWhenUnplaced)(_types2.default.callExpression(_types2.default.memberExpression(node, _types2.default.identifier("bind")), [object]), (0, _astLocUtils.getLoc)(object));
      }

      path.replaceWith((0, _astLocUtils.placeAtNode)(_types2.default.assignmentExpression("=", (0, _astLocUtils.placeAtLoc)(_types2.default.memberExpression(object, path.node.id), (0, _astLocUtils.getSurroundingLoc)([object, path.node.id])), node), node));
    }, ArrowFunctionExpression(path) {
      if (path.node.skinny) {
        (0, _functions.replaceWithPlainFunction)(path);
      } else if (path.node.generator) {
        (0, _functions.replaceWithBoundFunction)(path);
      }
    }, ClassBody(path) {
      var fatArrows = [],
          fatStaticArrows = [],
          constructorPath = void 0;
      path.node.body.forEach(function (method, i) {
        if (!_types2.default.isMethod(method)) return;

        if (method.kind === "constructor") {
          return constructorPath = path.get(`body.${i}.body`);
        } else if (method.static && method.skinny === false) {
          fatStaticArrows.push(method.key);
          return method.skinny = true; // prevent infinite recursion
        } else if (method.skinny === false) {
          fatArrows.push(method.key);
          return method.skinny = true; // prevent infinite recursion
        }
      });

      var maybeAddSuper = path.parentPath.node.superClass && constructorPath;
      if (fatArrows.length || maybeAddSuper) {
        constructorPath = (0, _classes.ensureConstructor)(path.parentPath, constructorPath, true);
      }

      if (fatArrows.length) {
        (0, _classes.bindMethodsInConstructor)(path.parentPath, constructorPath, fatArrows);
      }

      if (fatStaticArrows.length) {
        (0, _classes.bindMethods)(path.parentPath, fatStaticArrows);
      }
    }, JSXExpressionContainer(path) {
      (0, _comprehension.maybeTransformJSXExpressionWithSpreadLoop)(path);
    }, ArrayExpression(path) {
      (0, _comprehension.maybeTransformArrayWithSpreadLoops)(path);
    }, ObjectExpression(path) {
      var fatArrowKeys = function () {
        var _arr = [];
        for (var _arr3 = path.node.properties, _i2 = 0, _len2 = _arr3.length; _i2 < _len2; _i2++) {
          var _prop = _arr3[_i2];
          if ((0, _types.isa)(_prop, "Method") && _prop.skinny === false) {
            _prop.skinny = true; // prevent recursion
            _arr.push(_prop.key);
          }
        }return _arr;
      }();

      // ObjectExpression may get replaced with an IIFE here...
      if ((0, _comprehension.maybeTransformObjectWithSpreadLoops)(path)) path.resync();

      if (fatArrowKeys.length) (0, _classes.bindMethods)(path, fatArrowKeys);
    }, SpreadElement(path) {
      if (!path.node._safe) {
        // prevent infinite recursion
        (0, _safe.transformSafeSpreadElement)(path);
      }
    }, Function: {
      enter(path) {
        if ((0, _types.isa)(path.node, "FunctionDeclaration")) {
          // This is needed because named arrow declarations are new in
          // Lightscript and therefore not acknowledged by Babel's default
          // traversal algorithm for assessing declarations. We must
          // register the declaration by hand here.
          path.getFunctionParent().scope.registerDeclaration(path);
        }
      }, exit(path) {
        (0, _functions.addImplicitReturns)(path);
      } },

    IfExpression(path) {
      // TODO: better source mapping here -- place implicit null at end of consequent?
      var consequent = (0, _blocks.blockToExpression)(path.get("consequent"));
      var undef = path.scope.buildUndefinedNode();

      var alternate = path.node.alternate ? (0, _blocks.blockToExpression)(path.get("alternate")) : (0, _astLocUtils.placeAtNode)(undef, path.node);

      path.replaceWith((0, _astLocUtils.placeAtNode)(_types2.default.conditionalExpression(path.node.test, consequent, alternate), path.node));
    }, AssignmentExpression(path) {
      if (path.node.operator === "<-" || path.node.operator === "<!-") {
        path.node.operator = "=";
      }if ((0, _types.is)("MemberExpression", path.node.left)) return;

      if (path.node.isNowAssign === false) {
        throw path.buildCodeFrameError("Incorrect assignment: to reassign, use `now`; to assign as `const`, put on its own line.");
      }
    }, SafeAwaitExpression(path) {
      var errId = path.scope.generateUidIdentifier("err");
      var awaitIife = (0, _blocks.iife)((0, _astLocUtils.placeTreeAtLocWhenUnplaced)(_types2.default.tryStatement(_types2.default.blockStatement([_types2.default.returnStatement(_types2.default.awaitExpression(path.node.argument))]), _types2.default.catchClause(errId, _types2.default.blockStatement([_types2.default.returnStatement(errId)]))), (0, _astLocUtils.getLoc)(path.node)), true);

      // TODO: returntype annotation
      var awaitExpr = (0, _astLocUtils.placeTreeAtLocWhenUnplaced)(_types2.default.awaitExpression(awaitIife), (0, _astLocUtils.getLoc)(path.node));
      path.replaceWith(awaitExpr);
    }, AwaitExpression(path) {
      if ((0, _types.isa)(path.get("argument"), "ArrayExpression")) {
        var promiseDotAllCall = _types2.default.callExpression(_types2.default.memberExpression(_types2.default.identifier("Promise"), _types2.default.identifier("all")), [path.node.argument]);
        path.get("argument").replaceWith(promiseDotAllCall);
      }
    }, VariableDeclaration(path) {
      // Error on auto-const when shadowing variable
      if (path.node.kind === "const") {
        if (path.node.extra && path.node.extra.implicit === true) {
          var id = (0, _variables.getShadowingIdentifiers)(path)[0];
          if (id) {
            throw path.buildCodeFrameError(`\`${id.name}\` is shadowed from a higher scope. ` + `If you want to reassign the variable, use \`now ${id.name} = ...\`. ` + "If you want to declare a new shadowed \`const\` variable, " + `you must use \`const ${id.name} = ...\` explicitly.`);
          }
        }
      }
    } // collect functions to be imported for the stdlib
    , ReferencedIdentifier(path) {
      if (!(0, _variables.checkFalsePositiveReference)(path)) {
        (0, _stdlib.markIdentifier)(compilerState.stdlib, path);
      }
    }, ExistentialExpression(path) {
      (0, _safe.transformExistentialExpression)(path);
    }, MatchExpression(path) {
      matching.transformMatchExpression(path, opts.__linter);
    }, MatchStatement(path) {
      matching.transformMatchStatement(path, opts.__linter);
    }, CatchExpression(path) {
      (0, _catchExpression.transformCatchExpression)(path, opts.__linter);
    }, ForOfStatement(path) {
      // Auto-const
      var node = path.node;
      var left = node.left;

      if (left.type !== "VariableDeclaration" && !left.isNowAssign) {
        path.get("left").replaceWith((0, _astLocUtils.placeAtNode)(_types2.default.variableDeclaration("const", [(0, _astLocUtils.placeAtNode)(_types2.default.variableDeclarator(left), left)]), left));
      }
    } });
};

module.exports = exports["default"];